/*
Copyright 2020-2023 CAcert Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package handlers

import (
	"bytes"
	"encoding/base64"
	"encoding/json"
	"errors"
	"fmt"
	"net/http"
	"net/url"

	"github.com/lestrrat-go/jwx/jwk"
	"github.com/lestrrat-go/jwx/jwt"
	"github.com/lestrrat-go/jwx/jwt/openid"
	log "github.com/sirupsen/logrus"
	"golang.org/x/oauth2"

	"code.cacert.org/cacert/oidc-demo-app/internal/models"
	"code.cacert.org/cacert/oidc-demo-app/internal/services"
)

const (
	sessionName = "resource_session"

	oauth2RedirectStateLength = 8
)

func Authenticate(logger *log.Logger, oauth2Config *oauth2.Config, clientID string) func(http.Handler) http.Handler {
	return func(next http.Handler) http.Handler {
		return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
			session, err := services.GetSessionStore().Get(r, sessionName)
			if err != nil {
				http.Error(w, err.Error(), http.StatusInternalServerError)

				return
			}

			if _, ok := session.Values[sessionKeyIDToken]; ok {
				next.ServeHTTP(w, r)

				return
			}

			session.Values[sessionRedirectTarget] = r.URL.String()

			if err = session.Save(r, w); err != nil {
				http.Error(w, err.Error(), http.StatusInternalServerError)

				return
			}

			var authURL *url.URL

			if authURL, err = url.Parse(oauth2Config.Endpoint.AuthURL); err != nil {
				http.Error(w, err.Error(), http.StatusInternalServerError)

				return
			}

			queryValues := authURL.Query()
			queryValues.Set("client_id", clientID)
			queryValues.Set("response_type", "code")
			queryValues.Set("scope", "openid offline_access profile email")
			queryValues.Set("state", base64.URLEncoding.EncodeToString(services.GenerateKey(oauth2RedirectStateLength)))
			queryValues.Set("claims", getRequestedClaims(logger))
			authURL.RawQuery = queryValues.Encode()

			w.Header().Set("Location", authURL.String())
			w.WriteHeader(http.StatusFound)
		})
	}
}

func getRequestedClaims(logger *log.Logger) string {
	claims := make(models.OIDCClaimsRequest)
	claims["userinfo"] = make(models.ClaimElement)
	essentialItem := make(models.IndividualClaimRequest)
	essentialItem["essential"] = true
	claims["userinfo"]["https://auth.cacert.org/groups"] = &essentialItem

	target := make([]byte, 0)
	buf := bytes.NewBuffer(target)
	enc := json.NewEncoder(buf)

	if err := enc.Encode(claims); err != nil {
		logger.WithError(err).Warn("could not encode claims request parameter")
	}

	return buf.String()
}

func ParseIDToken(token string, keySet jwk.Set) (openid.Token, error) {
	var (
		parsedIDToken jwt.Token
		err           error
	)

	if parsedIDToken, err = jwt.ParseString(token, jwt.WithKeySet(keySet), jwt.WithToken(openid.New())); err != nil {
		return nil, fmt.Errorf("could not parse ID token: %w", err)
	}

	if v, ok := parsedIDToken.(openid.Token); ok {
		return v, nil
	}

	return nil, errors.New("ID token is no OpenID Connect Identity Token")
}
