/*
Copyright 2020-2023 CAcert Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package handlers

import (
	"fmt"
	"html/template"
	"io/fs"
	"net/http"
	"net/url"

	"github.com/lestrrat-go/jwx/jwk"
	"github.com/nicksnyder/go-i18n/v2/i18n"

	"code.cacert.org/cacert/oidc-demo-app/internal/services"
)

type IndexHandler struct {
	bundle         *i18n.Bundle
	indexTemplate  *template.Template
	keySet         jwk.Set
	logoutURL      string
	messageCatalog *services.MessageCatalog
	publicURL      string
}

func (h *IndexHandler) ServeHTTP(writer http.ResponseWriter, request *http.Request) {
	if request.Method != http.MethodGet {
		http.Error(writer, http.StatusText(http.StatusMethodNotAllowed), http.StatusMethodNotAllowed)

		return
	}

	if request.URL.Path != "/" {
		http.NotFound(writer, request)

		return
	}

	accept := request.Header.Get("Accept-Language")
	localizer := i18n.NewLocalizer(h.bundle, accept)

	writer.WriteHeader(http.StatusOK)

	session, err := services.GetSessionStore().Get(request, sessionName)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)

		return
	}

	logoutURL, err := url.Parse(h.logoutURL)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)

		return
	}

	var (
		idToken string
		ok      bool
	)

	if idToken, ok = session.Values[sessionKeyIDToken].(string); ok {
		logoutURL.RawQuery = url.Values{
			"id_token_hint":            []string{idToken},
			"post_logout_redirect_uri": []string{fmt.Sprintf("%s/after-logout", h.publicURL)},
		}.Encode()
	} else {
		return
	}

	oidcToken, err := ParseIDToken(idToken, h.keySet)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)

		return
	}

	writer.Header().Add("Content-Type", "text/html")

	msgLookup := h.messageCatalog.LookupMessage

	err = h.indexTemplate.Lookup("base").Execute(writer, map[string]interface{}{
		"Title": msgLookup("IndexTitle", nil, localizer),
		"Greeting": msgLookup("IndexGreeting", map[string]interface{}{
			"User": oidcToken.Name(),
		}, localizer),
		"IntroductionText": msgLookup("IndexIntroductionText", nil, localizer),
		"LogoutLabel":      msgLookup("LogoutLabel", nil, localizer),
		"LogoutURL":        logoutURL.String(),
	})
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)

		return
	}
}

func NewIndexHandler(
	bundle *i18n.Bundle,
	catalog *services.MessageCatalog,
	templateFS fs.FS,
	oidcInfo *services.OIDCInformation,
	publicURL string,
) (*IndexHandler, error) {
	indexTemplate, err := template.ParseFS(
		templateFS,
		"templates/base.gohtml", "templates/index.gohtml")
	if err != nil {
		return nil, fmt.Errorf("could not parse templates: %w", err)
	}

	return &IndexHandler{
		bundle:         bundle,
		indexTemplate:  indexTemplate,
		keySet:         oidcInfo.KeySet,
		logoutURL:      oidcInfo.OIDCConfiguration.EndSessionEndpoint,
		messageCatalog: catalog,
		publicURL:      publicURL,
	}, nil
}
