/*
Copyright 2020-2023 CAcert Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package services

import (
	"errors"
	"fmt"

	"github.com/nicksnyder/go-i18n/v2/i18n"
	"golang.org/x/text/language"

	"code.cacert.org/cacert/oidc-demo-app/translations"

	"github.com/BurntSushi/toml"
	log "github.com/sirupsen/logrus"
)

func AddMessages(catalog *MessageCatalog) {
	messages := make(map[string]*i18n.Message)
	messages["IndexGreeting"] = &i18n.Message{
		ID:    "IndexGreeting",
		Other: "Hello {{ .User }}",
	}
	messages["IndexTitle"] = &i18n.Message{
		ID:    "IndexTitle",
		Other: "Welcome to the Demo application",
	}
	messages["LogoutLabel"] = &i18n.Message{
		ID:          "LogoutLabel",
		Description: "A label on a logout button or link",
		Other:       "Logout",
	}
	messages["IndexIntroductionText"] = &i18n.Message{
		ID:    "IndexIntroductionText",
		Other: "This is an authorization protected resource",
	}

	catalog.AddMessages(messages)
}

type MessageCatalog struct {
	messages map[string]*i18n.Message
	logger   *log.Logger
}

func (m *MessageCatalog) AddMessages(messages map[string]*i18n.Message) {
	for key, value := range messages {
		m.messages[key] = value
	}
}

func (m *MessageCatalog) LookupErrorMessage(
	tag string,
	field string,
	value interface{},
	localizer *i18n.Localizer,
) string {
	fieldTag := fmt.Sprintf("%s-%s", field, tag)
	message, ok := m.messages[fieldTag]
	if !ok {
		m.logger.WithField("field_tag", fieldTag).Info("no specific error message for field and tag")

		message, ok = m.messages[tag]
		if !ok {
			m.logger.WithField("tag", tag).Info("no specific error message for tag")

			message, ok = m.messages["unknown"]
			if !ok {
				m.logger.Warn("no default translation found")

				return tag
			}
		}
	}

	translation, err := localizer.Localize(&i18n.LocalizeConfig{
		DefaultMessage: message,
		TemplateData: map[string]interface{}{
			"Value": value,
		},
	})
	if err != nil {
		m.logger.WithError(err).Error("localization failed")

		return tag
	}

	return translation
}

func (m *MessageCatalog) LookupMessage(
	id string,
	templateData map[string]interface{},
	localizer *i18n.Localizer,
) string {
	if message, ok := m.messages[id]; ok {
		translation, err := localizer.Localize(&i18n.LocalizeConfig{
			DefaultMessage: message,
			TemplateData:   templateData,
		})
		if err != nil {
			return m.handleLocalizeError(id, translation, err)
		}

		return translation
	}

	m.logger.WithField("id", id).Warn("no translation found for id")

	return id
}

func (m *MessageCatalog) handleLocalizeError(id string, translation string, err error) string {
	var messageNotFound *i18n.MessageNotFoundErr

	if errors.As(err, &messageNotFound) {
		m.logger.WithError(err).WithField("message", id).Warn("message not found")

		if translation != "" {
			return translation
		}
	} else {
		m.logger.WithError(err).WithField("message", id).Error("translation error")
	}

	return id
}

func InitI18n(logger *log.Logger, languages []string) (*i18n.Bundle, *MessageCatalog) {
	bundle := i18n.NewBundle(language.English)
	bundle.RegisterUnmarshalFunc("toml", toml.Unmarshal)

	for _, lang := range languages {
		bundleName := fmt.Sprintf("active.%s.toml", lang)

		bundleBytes, err := translations.Bundles.ReadFile(bundleName)
		if err != nil {
			logger.WithField("bundle", bundleName).Warn("message bundle not found")

			continue
		}

		bundle.MustParseMessageFileBytes(bundleBytes, bundleName)
	}

	catalog := initMessageCatalog(logger)

	return bundle, catalog
}

func initMessageCatalog(logger *log.Logger) *MessageCatalog {
	messages := make(map[string]*i18n.Message)
	messages["ErrorTitle"] = &i18n.Message{
		ID:    "ErrorTitle",
		Other: "An error has occurred",
	}

	return &MessageCatalog{messages: messages, logger: logger}
}
